"""Generated implementation of branch."""

# WARNING DO NOT EDIT
# This code was generated from branch.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..commit import Commit, CommitId
from ..user import UserId


@dataclasses.dataclass(frozen=True)
class BranchName:
    """Unique name of a branch.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BranchName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BranchName:
        """Validate and parse JSON data into an instance of BranchName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BranchName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BranchName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing BranchName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> BranchName:
        """Parse a JSON string such as a dictionary key."""
        return BranchName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class BranchRequest:
    """Request to create a new branch.
    
    Args:
        name (str): A data field.
        commit (CommitId): A data field.
    """
    
    name: str
    commit: CommitId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BranchRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string"
                },
                "commit": CommitId.json_schema()
            },
            "required": [
                "name",
                "commit",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BranchRequest:
        """Validate and parse JSON data into an instance of BranchRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BranchRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BranchRequest(
                name=str(data["name"]),
                commit=CommitId.from_json(data["commit"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BranchRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": str(self.name),
            "commit": self.commit.to_json()
        }


@dataclasses.dataclass(frozen=True)
class BranchUpdateRequest:
    """Request to update a branch to point to a new commit.
    
    Args:
        commit (CommitId): A data field.
        force (typing.Optional[bool]): A data field.
    """
    
    commit: CommitId
    force: typing.Optional[bool]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BranchUpdateRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "commit": CommitId.json_schema(),
                "force": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "boolean"},
                    ]
                }
            },
            "required": [
                "commit",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BranchUpdateRequest:
        """Validate and parse JSON data into an instance of BranchUpdateRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BranchUpdateRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BranchUpdateRequest(
                commit=CommitId.from_json(data["commit"]),
                force=(lambda v: v and bool(v))(data.get("force", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BranchUpdateRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "commit": self.commit.to_json(),
            "force": (lambda v: v and v)(self.force)
        }


@dataclasses.dataclass(frozen=True)
class BranchStatistics:
    """Commits ahead and behind the official branch.
    
    Args:
        commitsAhead (int): A data field.
        commitsBehind (int): A data field.
    """
    
    commitsAhead: int
    commitsBehind: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BranchStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "commitsAhead": {
                    "type": "integer"
                },
                "commitsBehind": {
                    "type": "integer"
                }
            },
            "required": [
                "commitsAhead",
                "commitsBehind",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BranchStatistics:
        """Validate and parse JSON data into an instance of BranchStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BranchStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BranchStatistics(
                commitsAhead=int(data["commitsAhead"]),
                commitsBehind=int(data["commitsBehind"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BranchStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "commitsAhead": int(self.commitsAhead),
            "commitsBehind": int(self.commitsBehind)
        }


@dataclasses.dataclass(frozen=True)
class Branch:
    """Information regarding a branch.
    
    Args:
        name (BranchName): A data field.
        created (typing.Optional[datetime.datetime]): A data field.
        author (typing.Optional[UserId]): A data field.
        head (Commit): A data field.
        statistics (typing.Optional[BranchStatistics]): A data field.
    """
    
    name: BranchName
    created: typing.Optional[datetime.datetime]
    author: typing.Optional[UserId]
    head: Commit
    statistics: typing.Optional[BranchStatistics]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Branch data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": BranchName.json_schema(),
                "created": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "author": {
                    "oneOf": [
                        {"type": "null"},
                        UserId.json_schema(),
                    ]
                },
                "head": Commit.json_schema(),
                "statistics": {
                    "oneOf": [
                        {"type": "null"},
                        BranchStatistics.json_schema(),
                    ]
                }
            },
            "required": [
                "name",
                "head",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Branch:
        """Validate and parse JSON data into an instance of Branch.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Branch.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Branch(
                name=BranchName.from_json(data["name"]),
                created=(lambda v: v and isodate.parse_datetime(v))(data.get("created", None)),
                author=(lambda v: v and UserId.from_json(v))(data.get("author", None)),
                head=Commit.from_json(data["head"]),
                statistics=(
                    lambda v: v and BranchStatistics.from_json(v)
                )(
                    data.get("statistics", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Branch",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": self.name.to_json(),
            "created": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.created),
            "author": (lambda v: v and v.to_json())(self.author),
            "head": self.head.to_json(),
            "statistics": (lambda v: v and v.to_json())(self.statistics)
        }



"""Generated implementation of event_store_creation_request."""

# WARNING DO NOT EDIT
# This code was generated from event-store-creation-request.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..access_control import AccessRule
from ..attribute import Attribute, SensitiveAttribute
from ..cluster import ClusterId
from ..event_store import EventStoreName, EventStoreTopicColumns, Topic
from ..label import Label
from ..schedule import Schedule


@dataclasses.dataclass(frozen=True)
class EventStoreCreationRequest:
    """Details for an event store.
    
    Args:
        name (EventStoreName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        accessRules (typing.List[AccessRule]): A data field.
        bootstrapServers (str): A data field.
        schemaRegistryUrl (str): A data field.
        kafkaPropertiesProviders (typing.List[SensitiveAttribute]): A data field.
        ingestions (typing.Dict[Topic, EventStoreTopicColumns]): A data field.
        connectBaseURI (typing.Optional[str]): A data field.
        scatterBaseURI (str): A data field.
        glacierBaseURI (str): A data field.
        batchIngestBaseURI (typing.Optional[str]): A data field.
        schedule (Schedule): A data field.
        cluster (ClusterId): A data field.
    """
    
    name: EventStoreName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    accessRules: typing.List[AccessRule]
    bootstrapServers: str
    schemaRegistryUrl: str
    kafkaPropertiesProviders: typing.List[SensitiveAttribute]
    ingestions: typing.Dict[Topic, EventStoreTopicColumns]
    connectBaseURI: typing.Optional[str]
    scatterBaseURI: str
    glacierBaseURI: str
    batchIngestBaseURI: typing.Optional[str]
    schedule: Schedule
    cluster: ClusterId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EventStoreCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": EventStoreName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "accessRules": {
                    "type": "array",
                    "item": AccessRule.json_schema()
                },
                "bootstrapServers": {
                    "type": "string"
                },
                "schemaRegistryUrl": {
                    "type": "string"
                },
                "kafkaPropertiesProviders": {
                    "type": "array",
                    "item": SensitiveAttribute.json_schema()
                },
                "ingestions": {
                    "type": "object",
                    "additionalProperties": EventStoreTopicColumns.json_schema()
                },
                "connectBaseURI": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "scatterBaseURI": {
                    "type": "string"
                },
                "glacierBaseURI": {
                    "type": "string"
                },
                "batchIngestBaseURI": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "schedule": Schedule.json_schema(),
                "cluster": ClusterId.json_schema()
            },
            "required": [
                "name",
                "description",
                "labels",
                "attributes",
                "accessRules",
                "bootstrapServers",
                "schemaRegistryUrl",
                "kafkaPropertiesProviders",
                "ingestions",
                "scatterBaseURI",
                "glacierBaseURI",
                "schedule",
                "cluster",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EventStoreCreationRequest:
        """Validate and parse JSON data into an instance of EventStoreCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EventStoreCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EventStoreCreationRequest(
                name=EventStoreName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                accessRules=[AccessRule.from_json(v) for v in data["accessRules"]],
                bootstrapServers=str(data["bootstrapServers"]),
                schemaRegistryUrl=str(data["schemaRegistryUrl"]),
                kafkaPropertiesProviders=[SensitiveAttribute.from_json(v) for v in data["kafkaPropertiesProviders"]],
                ingestions={
                    Topic.from_json_key(k): EventStoreTopicColumns.from_json(v) for k, v in data["ingestions"].items()
                },
                connectBaseURI=(lambda v: v and str(v))(data.get("connectBaseURI", None)),
                scatterBaseURI=str(data["scatterBaseURI"]),
                glacierBaseURI=str(data["glacierBaseURI"]),
                batchIngestBaseURI=(
                    lambda v: v and str(v)
                )(
                    data.get("batchIngestBaseURI", None)
                ),
                schedule=Schedule.from_json(data["schedule"]),
                cluster=ClusterId.from_json(data["cluster"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EventStoreCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "accessRules": [v.to_json() for v in self.accessRules],
            "bootstrapServers": str(self.bootstrapServers),
            "schemaRegistryUrl": str(self.schemaRegistryUrl),
            "kafkaPropertiesProviders": [v.to_json() for v in self.kafkaPropertiesProviders],
            "ingestions": {k.to_json_key(): v.to_json() for k, v in self.ingestions.items()},
            "connectBaseURI": (lambda v: v and str(v))(self.connectBaseURI),
            "scatterBaseURI": str(self.scatterBaseURI),
            "glacierBaseURI": str(self.glacierBaseURI),
            "batchIngestBaseURI": (lambda v: v and str(v))(self.batchIngestBaseURI),
            "schedule": self.schedule.to_json(),
            "cluster": self.cluster.to_json()
        }



"""Generated implementation of webhook."""

# WARNING DO NOT EDIT
# This code was generated from webhook.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


@dataclasses.dataclass(frozen=True)
class WebhookId:
    """Unique identifier of a webhook.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for WebhookId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> WebhookId:
        """Validate and parse JSON data into an instance of WebhookId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of WebhookId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return WebhookId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing WebhookId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> WebhookId:
        """Parse a JSON string such as a dictionary key."""
        return WebhookId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class WebhookName:
    """Unique name of a webhook.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for WebhookName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> WebhookName:
        """Validate and parse JSON data into an instance of WebhookName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of WebhookName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return WebhookName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing WebhookName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> WebhookName:
        """Parse a JSON string such as a dictionary key."""
        return WebhookName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class WebhookVersionId:
    """Unique identifier of a specific version of a webhook.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for WebhookVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> WebhookVersionId:
        """Validate and parse JSON data into an instance of WebhookVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of WebhookVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return WebhookVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing WebhookVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> WebhookVersionId:
        """Parse a JSON string such as a dictionary key."""
        return WebhookVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class EventConfiguration:
    """Webhook configuration for a specific event type."""
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EventConfiguration data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
            
            },
            "required": []
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EventConfiguration:
        """Validate and parse JSON data into an instance of EventConfiguration.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EventConfiguration.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EventConfiguration(
                
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EventConfiguration",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
        
        }


@dataclasses.dataclass(frozen=True)
class Webhook:
    """Configuration for a webhook.
    
    Args:
        id (WebhookId): A data field.
        name (WebhookName): A data field.
        description (str): A data field.
        url (str): A data field.
        mergeRequests (typing.Optional[EventConfiguration]): A data field.
        mergeRequestComments (typing.Optional[EventConfiguration]): A data field.
        commits (typing.Optional[EventConfiguration]): A data field.
        featureStoreRuns (typing.Optional[EventConfiguration]): A data field.
        monitoringRuns (typing.Optional[EventConfiguration]): A data field.
        cachingRuns (typing.Optional[EventConfiguration]): A data field.
        materialisationRuns (typing.Optional[EventConfiguration]): A data field.
        eventStoreRuns (typing.Optional[EventConfiguration]): A data field.
        version (WebhookVersionId): A data field.
    """
    
    id: WebhookId
    name: WebhookName
    description: str
    url: str
    mergeRequests: typing.Optional[EventConfiguration]
    mergeRequestComments: typing.Optional[EventConfiguration]
    commits: typing.Optional[EventConfiguration]
    featureStoreRuns: typing.Optional[EventConfiguration]
    monitoringRuns: typing.Optional[EventConfiguration]
    cachingRuns: typing.Optional[EventConfiguration]
    materialisationRuns: typing.Optional[EventConfiguration]
    eventStoreRuns: typing.Optional[EventConfiguration]
    version: WebhookVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Webhook data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": WebhookId.json_schema(),
                "name": WebhookName.json_schema(),
                "description": {
                    "type": "string"
                },
                "url": {
                    "type": "string"
                },
                "mergeRequests": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "mergeRequestComments": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "commits": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "featureStoreRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "monitoringRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "cachingRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "materialisationRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "eventStoreRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "version": WebhookVersionId.json_schema()
            },
            "required": [
                "id",
                "name",
                "description",
                "url",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Webhook:
        """Validate and parse JSON data into an instance of Webhook.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Webhook.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Webhook(
                id=WebhookId.from_json(data["id"]),
                name=WebhookName.from_json(data["name"]),
                description=str(data["description"]),
                url=str(data["url"]),
                mergeRequests=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("mergeRequests", None)
                ),
                mergeRequestComments=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("mergeRequestComments", None)
                ),
                commits=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("commits", None)
                ),
                featureStoreRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("featureStoreRuns", None)
                ),
                monitoringRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("monitoringRuns", None)
                ),
                cachingRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("cachingRuns", None)
                ),
                materialisationRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("materialisationRuns", None)
                ),
                eventStoreRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("eventStoreRuns", None)
                ),
                version=WebhookVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Webhook",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "url": str(self.url),
            "mergeRequests": (lambda v: v and v.to_json())(self.mergeRequests),
            "mergeRequestComments": (lambda v: v and v.to_json())(self.mergeRequestComments),
            "commits": (lambda v: v and v.to_json())(self.commits),
            "featureStoreRuns": (lambda v: v and v.to_json())(self.featureStoreRuns),
            "monitoringRuns": (lambda v: v and v.to_json())(self.monitoringRuns),
            "cachingRuns": (lambda v: v and v.to_json())(self.cachingRuns),
            "materialisationRuns": (lambda v: v and v.to_json())(self.materialisationRuns),
            "eventStoreRuns": (lambda v: v and v.to_json())(self.eventStoreRuns),
            "version": self.version.to_json()
        }


@dataclasses.dataclass(frozen=True)
class WebhookCreationRequest:
    """Request to create or update a webhook.
    
    Args:
        name (WebhookName): A data field.
        description (str): A data field.
        url (str): A data field.
        mergeRequests (typing.Optional[EventConfiguration]): A data field.
        mergeRequestComments (typing.Optional[EventConfiguration]): A data field.
        commits (typing.Optional[EventConfiguration]): A data field.
        featureStoreRuns (typing.Optional[EventConfiguration]): A data field.
        monitoringRuns (typing.Optional[EventConfiguration]): A data field.
        cachingRuns (typing.Optional[EventConfiguration]): A data field.
        materialisationRuns (typing.Optional[EventConfiguration]): A data field.
        eventStoreRuns (typing.Optional[EventConfiguration]): A data field.
    """
    
    name: WebhookName
    description: str
    url: str
    mergeRequests: typing.Optional[EventConfiguration]
    mergeRequestComments: typing.Optional[EventConfiguration]
    commits: typing.Optional[EventConfiguration]
    featureStoreRuns: typing.Optional[EventConfiguration]
    monitoringRuns: typing.Optional[EventConfiguration]
    cachingRuns: typing.Optional[EventConfiguration]
    materialisationRuns: typing.Optional[EventConfiguration]
    eventStoreRuns: typing.Optional[EventConfiguration]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for WebhookCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": WebhookName.json_schema(),
                "description": {
                    "type": "string"
                },
                "url": {
                    "type": "string"
                },
                "mergeRequests": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "mergeRequestComments": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "commits": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "featureStoreRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "monitoringRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "cachingRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "materialisationRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                },
                "eventStoreRuns": {
                    "oneOf": [
                        {"type": "null"},
                        EventConfiguration.json_schema(),
                    ]
                }
            },
            "required": [
                "name",
                "description",
                "url",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> WebhookCreationRequest:
        """Validate and parse JSON data into an instance of WebhookCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of WebhookCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return WebhookCreationRequest(
                name=WebhookName.from_json(data["name"]),
                description=str(data["description"]),
                url=str(data["url"]),
                mergeRequests=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("mergeRequests", None)
                ),
                mergeRequestComments=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("mergeRequestComments", None)
                ),
                commits=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("commits", None)
                ),
                featureStoreRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("featureStoreRuns", None)
                ),
                monitoringRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("monitoringRuns", None)
                ),
                cachingRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("cachingRuns", None)
                ),
                materialisationRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("materialisationRuns", None)
                ),
                eventStoreRuns=(
                    lambda v: v and EventConfiguration.from_json(v)
                )(
                    data.get("eventStoreRuns", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing WebhookCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": self.name.to_json(),
            "description": str(self.description),
            "url": str(self.url),
            "mergeRequests": (lambda v: v and v.to_json())(self.mergeRequests),
            "mergeRequestComments": (lambda v: v and v.to_json())(self.mergeRequestComments),
            "commits": (lambda v: v and v.to_json())(self.commits),
            "featureStoreRuns": (lambda v: v and v.to_json())(self.featureStoreRuns),
            "monitoringRuns": (lambda v: v and v.to_json())(self.monitoringRuns),
            "cachingRuns": (lambda v: v and v.to_json())(self.cachingRuns),
            "materialisationRuns": (lambda v: v and v.to_json())(self.materialisationRuns),
            "eventStoreRuns": (lambda v: v and v.to_json())(self.eventStoreRuns)
        }


@dataclasses.dataclass(frozen=True)
class WebhookLog:
    """Webhook delivery attempt log record.
    
    Args:
        id (int): A data field.
        hook (WebhookId): A data field.
        payload (str): A data field.
        url (str): A data field.
        responseCode (int): A data field.
        responseBody (str): A data field.
        delivered (datetime.datetime): A data field.
    """
    
    id: int
    hook: WebhookId
    payload: str
    url: str
    responseCode: int
    responseBody: str
    delivered: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for WebhookLog data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": {
                    "type": "integer"
                },
                "hook": WebhookId.json_schema(),
                "payload": {
                    "type": "string"
                },
                "url": {
                    "type": "string"
                },
                "responseCode": {
                    "type": "integer"
                },
                "responseBody": {
                    "type": "string"
                },
                "delivered": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "hook",
                "payload",
                "url",
                "responseCode",
                "responseBody",
                "delivered",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> WebhookLog:
        """Validate and parse JSON data into an instance of WebhookLog.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of WebhookLog.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return WebhookLog(
                id=int(data["id"]),
                hook=WebhookId.from_json(data["hook"]),
                payload=str(data["payload"]),
                url=str(data["url"]),
                responseCode=int(data["responseCode"]),
                responseBody=str(data["responseBody"]),
                delivered=isodate.parse_datetime(data["delivered"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing WebhookLog",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id,
            "hook": self.hook.to_json(),
            "payload": str(self.payload),
            "url": str(self.url),
            "responseCode": int(self.responseCode),
            "responseBody": str(self.responseBody),
            "delivered": self.delivered.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class WebhookMessage:
    """A message to be send to a webhook.
    
    Args:
        id (int): A data field.
        hook (WebhookId): A data field.
        payload (str): A data field.
        created (datetime.datetime): A data field.
        backoff (int): A data field.
    """
    
    id: int
    hook: WebhookId
    payload: str
    created: datetime.datetime
    backoff: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for WebhookMessage data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": {
                    "type": "integer"
                },
                "hook": WebhookId.json_schema(),
                "payload": {
                    "type": "string"
                },
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "backoff": {
                    "type": "integer"
                }
            },
            "required": [
                "id",
                "hook",
                "payload",
                "created",
                "backoff",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> WebhookMessage:
        """Validate and parse JSON data into an instance of WebhookMessage.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of WebhookMessage.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return WebhookMessage(
                id=int(data["id"]),
                hook=WebhookId.from_json(data["hook"]),
                payload=str(data["payload"]),
                created=isodate.parse_datetime(data["created"]),
                backoff=int(data["backoff"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing WebhookMessage",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id,
            "hook": self.hook.to_json(),
            "payload": str(self.payload),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "backoff": int(self.backoff)
        }



"""Generated implementation of scheduler."""

# WARNING DO NOT EDIT
# This code was generated from scheduler.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..date_range import DateRange
from ..event_store import EventStoreId
from ..feature_store import FeatureStoreId
from ..jobs import TableMonitoringJobId
from ..source_reference import SourceReference
from ..table_caching import TableCachingJobId
from ..view_materialisation import ViewMaterialisationJobId


@dataclasses.dataclass(frozen=True)
class SchedulerRunRequest:
    """A request for the scheduler to immediately kick off a job.
    
    Only one job id should be set in each call to the
    endpoint.
    
    If the job is an event store batch run, then also include
    a subject and source.
    
    Args:
        featureStoreId (typing.Optional[FeatureStoreId]): A data field.
        tableMonitoringJobId (typing.Optional[TableMonitoringJobId]): A data field.
        tableCachingJobId (typing.Optional[TableCachingJobId]): A data field.
        viewMaterialisationJobId (typing.Optional[ViewMaterialisationJobId]): A data field.
        eventStoreId (typing.Optional[EventStoreId]): A data field.
        dateRange (typing.Optional[DateRange]): A data field.
        subject (typing.Optional[str]): A data field.
        source (typing.Optional[SourceReference]): A data field.
    """
    
    featureStoreId: typing.Optional[FeatureStoreId]
    tableMonitoringJobId: typing.Optional[TableMonitoringJobId]
    tableCachingJobId: typing.Optional[TableCachingJobId]
    viewMaterialisationJobId: typing.Optional[ViewMaterialisationJobId]
    eventStoreId: typing.Optional[EventStoreId]
    dateRange: typing.Optional[DateRange]
    subject: typing.Optional[str]
    source: typing.Optional[SourceReference]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for SchedulerRunRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "featureStoreId": {
                    "oneOf": [
                        {"type": "null"},
                        FeatureStoreId.json_schema(),
                    ]
                },
                "tableMonitoringJobId": {
                    "oneOf": [
                        {"type": "null"},
                        TableMonitoringJobId.json_schema(),
                    ]
                },
                "tableCachingJobId": {
                    "oneOf": [
                        {"type": "null"},
                        TableCachingJobId.json_schema(),
                    ]
                },
                "viewMaterialisationJobId": {
                    "oneOf": [
                        {"type": "null"},
                        ViewMaterialisationJobId.json_schema(),
                    ]
                },
                "eventStoreId": {
                    "oneOf": [
                        {"type": "null"},
                        EventStoreId.json_schema(),
                    ]
                },
                "dateRange": {
                    "oneOf": [
                        {"type": "null"},
                        DateRange.json_schema(),
                    ]
                },
                "subject": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "source": {
                    "oneOf": [
                        {"type": "null"},
                        SourceReference.json_schema(),
                    ]
                }
            },
            "required": []
        }
    
    @classmethod
    def from_json(cls, data: dict) -> SchedulerRunRequest:
        """Validate and parse JSON data into an instance of SchedulerRunRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of SchedulerRunRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return SchedulerRunRequest(
                featureStoreId=(
                    lambda v: v and FeatureStoreId.from_json(v)
                )(
                    data.get("featureStoreId", None)
                ),
                tableMonitoringJobId=(
                    lambda v: v and TableMonitoringJobId.from_json(v)
                )(
                    data.get("tableMonitoringJobId", None)
                ),
                tableCachingJobId=(
                    lambda v: v and TableCachingJobId.from_json(v)
                )(
                    data.get("tableCachingJobId", None)
                ),
                viewMaterialisationJobId=(
                    lambda v: v and ViewMaterialisationJobId.from_json(v)
                )(
                    data.get("viewMaterialisationJobId", None)
                ),
                eventStoreId=(
                    lambda v: v and EventStoreId.from_json(v)
                )(
                    data.get("eventStoreId", None)
                ),
                dateRange=(lambda v: v and DateRange.from_json(v))(data.get("dateRange", None)),
                subject=(lambda v: v and str(v))(data.get("subject", None)),
                source=(lambda v: v and SourceReference.from_json(v))(data.get("source", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing SchedulerRunRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "featureStoreId": (lambda v: v and v.to_json())(self.featureStoreId),
            "tableMonitoringJobId": (lambda v: v and v.to_json())(self.tableMonitoringJobId),
            "tableCachingJobId": (lambda v: v and v.to_json())(self.tableCachingJobId),
            "viewMaterialisationJobId": (lambda v: v and v.to_json())(self.viewMaterialisationJobId),
            "eventStoreId": (lambda v: v and v.to_json())(self.eventStoreId),
            "dateRange": (lambda v: v and v.to_json())(self.dateRange),
            "subject": (lambda v: v and str(v))(self.subject),
            "source": (lambda v: v and v.to_json())(self.source)
        }



"""Generated implementation of cluster_creation_request."""

# WARNING DO NOT EDIT
# This code was generated from cluster-creation-request.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..attribute import Attribute
from ..cluster import ClusterName, SparkConfig
from ..credentials_provider_config import CredentialsProviderConfig
from ..label import Label


@dataclasses.dataclass(frozen=True)
class ClusterCreationRequest(abc.ABC):
    """Request to create a new cluster configuration.
    
    Args:
        attributes (typing.List[Attribute]): A data field.
        description (str): A data field.
        isPreviewCluster (bool): A data field.
        labels (typing.List[Label]): A data field.
        name (ClusterName): A data field.
        sparkConfig (SparkConfig): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    attributes: typing.List[Attribute]
    description: str
    isPreviewCluster: bool
    labels: typing.List[Label]
    name: ClusterName
    sparkConfig: SparkConfig
    
    @classmethod
    def json_schema(cls) -> ClusterCreationRequest:
        """JSON schema for variant ClusterCreationRequest.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ClusterCreationRequest:
        """Validate and parse JSON data into an instance of ClusterCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ClusterCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ClusterCreationRequest", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class LocalClusterCreationRequest(ClusterCreationRequest):
    """Request to create a new local cluster configuration.
    
    Args:
        name (ClusterName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        isPreviewCluster (bool): A data field.
        anamlServerUrl (str): A data field.
        credentialsProvider (CredentialsProviderConfig): A data field.
        sparkConfig (SparkConfig): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "local"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    name: ClusterName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    isPreviewCluster: bool
    anamlServerUrl: str
    credentialsProvider: CredentialsProviderConfig
    sparkConfig: SparkConfig
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for LocalClusterCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "name": ClusterName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "isPreviewCluster": {
                    "type": "boolean"
                },
                "anamlServerUrl": {
                    "type": "string"
                },
                "credentialsProvider": CredentialsProviderConfig.json_schema(),
                "sparkConfig": SparkConfig.json_schema()
            },
            "required": [
                "adt_type",
                "name",
                "description",
                "labels",
                "attributes",
                "isPreviewCluster",
                "anamlServerUrl",
                "credentialsProvider",
                "sparkConfig",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> LocalClusterCreationRequest:
        """Validate and parse JSON data into an instance of LocalClusterCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of LocalClusterCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return LocalClusterCreationRequest(
                name=ClusterName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                isPreviewCluster=bool(data["isPreviewCluster"]),
                anamlServerUrl=str(data["anamlServerUrl"]),
                credentialsProvider=CredentialsProviderConfig.from_json(data["credentialsProvider"]),
                sparkConfig=SparkConfig.from_json(data["sparkConfig"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing LocalClusterCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "isPreviewCluster": self.isPreviewCluster,
            "anamlServerUrl": str(self.anamlServerUrl),
            "credentialsProvider": self.credentialsProvider.to_json(),
            "sparkConfig": self.sparkConfig.to_json()
        }


@dataclasses.dataclass(frozen=True)
class SparkServerClusterCreationRequest(ClusterCreationRequest):
    """Request to create a new Spark Server cluster configuration.
    
    Args:
        name (ClusterName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        isPreviewCluster (bool): A data field.
        sparkServerUrl (str): A data field.
        sparkConfig (SparkConfig): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "sparkserver"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    name: ClusterName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    isPreviewCluster: bool
    sparkServerUrl: str
    sparkConfig: SparkConfig
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for SparkServerClusterCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "name": ClusterName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "isPreviewCluster": {
                    "type": "boolean"
                },
                "sparkServerUrl": {
                    "type": "string"
                },
                "sparkConfig": SparkConfig.json_schema()
            },
            "required": [
                "adt_type",
                "name",
                "description",
                "labels",
                "attributes",
                "isPreviewCluster",
                "sparkServerUrl",
                "sparkConfig",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> SparkServerClusterCreationRequest:
        """Validate and parse JSON data into an instance of SparkServerClusterCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of SparkServerClusterCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return SparkServerClusterCreationRequest(
                name=ClusterName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                isPreviewCluster=bool(data["isPreviewCluster"]),
                sparkServerUrl=str(data["sparkServerUrl"]),
                sparkConfig=SparkConfig.from_json(data["sparkConfig"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing SparkServerClusterCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "isPreviewCluster": self.isPreviewCluster,
            "sparkServerUrl": str(self.sparkServerUrl),
            "sparkConfig": self.sparkConfig.to_json()
        }



"""Generated implementation of preview_summary."""

# WARNING DO NOT EDIT
# This code was generated from preview-summary.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..entity import EntityName
from ..summary_statistics import SummaryStatistics


@dataclasses.dataclass(frozen=True)
class PreviewSummary:
    """Preview Summary for Features.
    
    Args:
        entityName (EntityName): A data field.
        statistics (typing.List[SummaryStatistics]): A data field.
    """
    
    entityName: EntityName
    statistics: typing.List[SummaryStatistics]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for PreviewSummary data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "entityName": EntityName.json_schema(),
                "statistics": {
                    "type": "array",
                    "item": SummaryStatistics.json_schema()
                }
            },
            "required": [
                "entityName",
                "statistics",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> PreviewSummary:
        """Validate and parse JSON data into an instance of PreviewSummary.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of PreviewSummary.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return PreviewSummary(
                entityName=EntityName.from_json(data["entityName"]),
                statistics=[SummaryStatistics.from_json(v) for v in data["statistics"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing PreviewSummary",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "entityName": self.entityName.to_json(),
            "statistics": [v.to_json() for v in self.statistics]
        }



"""Generated implementation of entity_mapping_creation_request."""

# WARNING DO NOT EDIT
# This code was generated from entity-mapping-creation-request.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..entity import EntityId
from ..feature_id import FeatureId


@dataclasses.dataclass(frozen=True)
class EntityMappingCreationRequest:
    """Request to create a new entity mapping.
    
    Args:
        from_ (EntityId): A data field.
        to (EntityId): A data field.
        mapping (FeatureId): A data field.
    """
    
    from_: EntityId
    to: EntityId
    mapping: FeatureId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityMappingCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "from": EntityId.json_schema(),
                "to": EntityId.json_schema(),
                "mapping": FeatureId.json_schema()
            },
            "required": [
                "from",
                "to",
                "mapping",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityMappingCreationRequest:
        """Validate and parse JSON data into an instance of EntityMappingCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityMappingCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityMappingCreationRequest(
                from_=EntityId.from_json(data["from"]),
                to=EntityId.from_json(data["to"]),
                mapping=FeatureId.from_json(data["mapping"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EntityMappingCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "from": self.from_.to_json(),
            "to": self.to.to_json(),
            "mapping": self.mapping.to_json()
        }



"""Generated implementation of user."""

# WARNING DO NOT EDIT
# This code was generated from user.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..roles import Role


@dataclasses.dataclass(frozen=True)
class UserId:
    """Unique identifier of a user, internal to Anaml.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserId:
        """Validate and parse JSON data into an instance of UserId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing UserId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> UserId:
        """Parse a JSON string such as a dictionary key."""
        return UserId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class UserEmail:
    """Unique email of a user.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserEmail data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserEmail:
        """Validate and parse JSON data into an instance of UserEmail.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserEmail.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserEmail(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing UserEmail", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> UserEmail:
        """Parse a JSON string such as a dictionary key."""
        return UserEmail(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class UserGroupId:
    """Unique identifier for a user group.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserGroupId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroupId:
        """Validate and parse JSON data into an instance of UserGroupId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroupId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing UserGroupId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> UserGroupId:
        """Parse a JSON string such as a dictionary key."""
        return UserGroupId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class User:
    """Details of a user account.
    
    Args:
        id (UserId): A data field.
        email (UserEmail): A data field.
        name (str): A data field.
        givenName (typing.Optional[str]): A data field.
        surname (typing.Optional[str]): A data field.
        lastLogin (typing.Optional[datetime.datetime]): A data field.
        created (typing.Optional[datetime.datetime]): A data field.
        modified (typing.Optional[datetime.datetime]): A data field.
        roles (typing.List[Role]): A data field.
        inheritedRoles (typing.List[Role]): A data field.
        groups (typing.List[UserGroupId]): A data field.
    """
    
    id: UserId
    email: UserEmail
    name: str
    givenName: typing.Optional[str]
    surname: typing.Optional[str]
    lastLogin: typing.Optional[datetime.datetime]
    created: typing.Optional[datetime.datetime]
    modified: typing.Optional[datetime.datetime]
    roles: typing.List[Role]
    inheritedRoles: typing.List[Role]
    groups: typing.List[UserGroupId]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for User data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": UserId.json_schema(),
                "email": UserEmail.json_schema(),
                "name": {
                    "type": "string"
                },
                "givenName": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "surname": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "lastLogin": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "created": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "modified": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "roles": {
                    "type": "array",
                    "item": Role.json_schema()
                },
                "inheritedRoles": {
                    "type": "array",
                    "item": Role.json_schema()
                },
                "groups": {
                    "type": "array",
                    "item": UserGroupId.json_schema()
                }
            },
            "required": [
                "id",
                "email",
                "name",
                "roles",
                "inheritedRoles",
                "groups",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> User:
        """Validate and parse JSON data into an instance of User.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of User.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return User(
                id=UserId.from_json(data["id"]),
                email=UserEmail.from_json(data["email"]),
                name=str(data["name"]),
                givenName=(lambda v: v and str(v))(data.get("givenName", None)),
                surname=(lambda v: v and str(v))(data.get("surname", None)),
                lastLogin=(
                    lambda v: v and isodate.parse_datetime(v)
                )(
                    data.get("lastLogin", None)
                ),
                created=(lambda v: v and isodate.parse_datetime(v))(data.get("created", None)),
                modified=(
                    lambda v: v and isodate.parse_datetime(v)
                )(
                    data.get("modified", None)
                ),
                roles=[Role.from_json(v) for v in data["roles"]],
                inheritedRoles=[Role.from_json(v) for v in data["inheritedRoles"]],
                groups=[UserGroupId.from_json(v) for v in data["groups"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing User",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "email": self.email.to_json(),
            "name": str(self.name),
            "givenName": (lambda v: v and str(v))(self.givenName),
            "surname": (lambda v: v and str(v))(self.surname),
            "lastLogin": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.lastLogin),
            "created": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.created),
            "modified": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.modified),
            "roles": [v.to_json() for v in self.roles],
            "inheritedRoles": [v.to_json() for v in self.inheritedRoles],
            "groups": [v.to_json() for v in self.groups]
        }


@dataclasses.dataclass(frozen=True)
class UserCreationRequest:
    """Request to create a new user account.
    
    Args:
        email (UserEmail): A data field.
        password (typing.Optional[str]): A data field.
        name (str): A data field.
        givenName (typing.Optional[str]): A data field.
        surname (typing.Optional[str]): A data field.
        roles (typing.List[Role]): A data field.
    """
    
    email: UserEmail
    password: typing.Optional[str]
    name: str
    givenName: typing.Optional[str]
    surname: typing.Optional[str]
    roles: typing.List[Role]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "email": UserEmail.json_schema(),
                "password": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "name": {
                    "type": "string"
                },
                "givenName": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "surname": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "roles": {
                    "type": "array",
                    "item": Role.json_schema()
                }
            },
            "required": [
                "email",
                "name",
                "roles",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserCreationRequest:
        """Validate and parse JSON data into an instance of UserCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserCreationRequest(
                email=UserEmail.from_json(data["email"]),
                password=(lambda v: v and str(v))(data.get("password", None)),
                name=str(data["name"]),
                givenName=(lambda v: v and str(v))(data.get("givenName", None)),
                surname=(lambda v: v and str(v))(data.get("surname", None)),
                roles=[Role.from_json(v) for v in data["roles"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing UserCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "email": self.email.to_json(),
            "password": (lambda v: v and str(v))(self.password),
            "name": str(self.name),
            "givenName": (lambda v: v and str(v))(self.givenName),
            "surname": (lambda v: v and str(v))(self.surname),
            "roles": [v.to_json() for v in self.roles]
        }


@dataclasses.dataclass(frozen=True)
class UserUpdateRequest:
    """Request to update a user account.
    
    Args:
        email (UserEmail): A data field.
        name (str): A data field.
        givenName (typing.Optional[str]): A data field.
        surname (typing.Optional[str]): A data field.
    """
    
    email: UserEmail
    name: str
    givenName: typing.Optional[str]
    surname: typing.Optional[str]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserUpdateRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "email": UserEmail.json_schema(),
                "name": {
                    "type": "string"
                },
                "givenName": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "surname": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                }
            },
            "required": [
                "email",
                "name",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserUpdateRequest:
        """Validate and parse JSON data into an instance of UserUpdateRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserUpdateRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserUpdateRequest(
                email=UserEmail.from_json(data["email"]),
                name=str(data["name"]),
                givenName=(lambda v: v and str(v))(data.get("givenName", None)),
                surname=(lambda v: v and str(v))(data.get("surname", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing UserUpdateRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "email": self.email.to_json(),
            "name": str(self.name),
            "givenName": (lambda v: v and str(v))(self.givenName),
            "surname": (lambda v: v and str(v))(self.surname)
        }


@dataclasses.dataclass(frozen=True)
class UserUpdateRequestWithRoles:
    """Request to update a user account and roles they are assigned.
    
    Args:
        email (UserEmail): A data field.
        name (str): A data field.
        givenName (typing.Optional[str]): A data field.
        surname (typing.Optional[str]): A data field.
        roles (typing.List[Role]): A data field.
    """
    
    email: UserEmail
    name: str
    givenName: typing.Optional[str]
    surname: typing.Optional[str]
    roles: typing.List[Role]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserUpdateRequestWithRoles data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "email": UserEmail.json_schema(),
                "name": {
                    "type": "string"
                },
                "givenName": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "surname": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "roles": {
                    "type": "array",
                    "item": Role.json_schema()
                }
            },
            "required": [
                "email",
                "name",
                "roles",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserUpdateRequestWithRoles:
        """Validate and parse JSON data into an instance of UserUpdateRequestWithRoles.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserUpdateRequestWithRoles.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserUpdateRequestWithRoles(
                email=UserEmail.from_json(data["email"]),
                name=str(data["name"]),
                givenName=(lambda v: v and str(v))(data.get("givenName", None)),
                surname=(lambda v: v and str(v))(data.get("surname", None)),
                roles=[Role.from_json(v) for v in data["roles"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing UserUpdateRequestWithRoles",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "email": self.email.to_json(),
            "name": str(self.name),
            "givenName": (lambda v: v and str(v))(self.givenName),
            "surname": (lambda v: v and str(v))(self.surname),
            "roles": [v.to_json() for v in self.roles]
        }



"""Generated implementation of select_expression."""

# WARNING DO NOT EDIT
# This code was generated from select-expression.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


@dataclasses.dataclass(frozen=True)
class SelectExpression:
    """SQL expression for select clause.
    
    Args:
        sql (str): A data field.
    """
    
    sql: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for SelectExpression data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "sql": {
                    "type": "string"
                }
            },
            "required": [
                "sql",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> SelectExpression:
        """Validate and parse JSON data into an instance of SelectExpression.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of SelectExpression.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return SelectExpression(
                sql=str(data["sql"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing SelectExpression",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "sql": str(self.sql)
        }



#
# Copyright 2021 Simple Machines Pty Ltd - All Rights Reserved.
#
# This file is part of Anaml.
#
# Unauthorized copying and/or distribution of this file, via any medium is
# strictly prohibited.
#
"""Interfaces and implementation of Anaml server serialisation."""
import dataclasses
import datetime
import enum
import logging
import uuid
from abc import ABC, abstractmethod
from typing import Optional, TypeVar, Type, Any

import isodate
from jsonschema import validate
from jsonschema.exceptions import ValidationError


# Type variables used in the serialisation interface.
S = TypeVar('S', bound='Serialisation')
E = TypeVar('E', bound='AnamlBaseEnum')
D = TypeVar('D', bound='AnamlBaseClass')


TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
INSTANT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


logger = logging.getLogger(__name__)


def json_safe(data: Any, datetime_format: Optional[str] = None) -> Any:
    """Convert values to JSON-safe values."""
    if datetime_format is None:
        datetime_format = INSTANT_FORMAT
    if isinstance(data, str) or isinstance(data, bool) or isinstance(data, float) or isinstance(data, int):
        return data
    elif data is None:
        return None
    elif isinstance(data, AnamlBaseClass):
        return data.to_json()
    elif isinstance(data, enum.Enum):
        if callable(getattr(data, 'to_json', None)):
            return data.to_json()
        else:
            return data.value
    elif isinstance(data, uuid.UUID):
        return str(data)
    elif isinstance(data, datetime.timedelta):
        return isodate.duration_isoformat(data)
    elif isinstance(data, datetime.time):
        return data.strftime("%H:%M:%S")
    elif isinstance(data, datetime.datetime):
        # NB: datetime <: date, so make sure you don't switch the order here!
        if datetime_format == INSTANT_FORMAT:
            return data.astimezone(datetime.timezone.utc).strftime(datetime_format)
        else:
            return data.strftime(datetime_format)
    elif isinstance(data, datetime.date):
        return data.isoformat()
    elif isinstance(data, dict):
        return {k: json_safe(v, datetime_format) for k, v in data.items()}
    elif isinstance(data, list):
        return [json_safe(v, datetime_format) for v in data]
    else:
        # TODO: This is probably an error.
        return data


class Serialisation:
    """Serialisation interface for Anaml data-types.

    We use this on dataclasses and enums to
    """

    @classmethod
    def json_schema(cls) -> Optional[dict]:
        """JSON Schema for this resource type."""
        return None

    @classmethod
    @abstractmethod
    def from_dict(cls: Type[S], data: dict) -> S:
        """Construct an instance from a validated dictionary of JSON data."""
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> dict:
        """Return a dictionary of fields and values from this object.

        Unlike :ref:`dataclasses.asdict`, this method does not decompose the field values.
        It should be the case that `obj == type(obj)(**self.to_dict())`.
        """
        raise NotImplementedError

    def to_json(self) -> dict:
        """Return a JSON dictionary encoding the object."""
        return json_safe(self.to_dict())

    @classmethod
    def from_json(cls: Type[S], d: dict) -> S:
        """Validate a dictionary of JSON data and parse it into a new instance of this type."""
        try:
            schema = cls.json_schema()
            if schema:
                validate(d, schema)
            return cls.from_dict(d)
        except ValidationError:
            logger.error("Unable to validate {klass} schema {schema} from dict {data}".format(
                klass=cls.__name__, schema=cls.json_schema(), data=d
            ))
            raise


@dataclasses.dataclass(frozen=True)
class AnamlBaseClass(Serialisation, ABC):
    """Base class for data types used in the Anaml API."""

    def to_dict(self) -> dict:
        """Return the object as a dictionary.

        Unlike :ref:`dataclasses.asdict`, this method is not recursive: it returns a dictionary
        of this object only. The values of this object's fields are included unchanged.
        """
        return {
            f.name: getattr(self, f.name) for f in dataclasses.fields(self)
        }


class AnamlBaseEnum(Serialisation, enum.Enum):
    """Base class for enumerations used in the Anaml API."""

    @classmethod
    def json_schema(cls) -> Optional[dict]:
        """Build JSON schema for enumerations encoded as an ADT."""
        return {
            "type": "object",
            "properties": {
                "adt_type": {"type": "string", "enum": [item.value for item in cls]}
            },
            "required": ["adt_type"]
        }

    @classmethod
    def from_dict(cls: Type[E], data: dict) -> E:
        """Parse an enumeration member from valid JSON data."""
        return cls(data['adt_type'])

    def to_dict(self) -> dict:
        """Return the object as a dictionary.

        Unlike :ref:`dataclasses.asdict`, this method is not recursive: it returns a dictionary
        of this object only. The values of this object's fields are included unchanged.
        """
        return {"adt_type": self.value}


class AnamlDirectEnum(AnamlBaseEnum):
    """Base class for enumerations that are directly encoded."""

    @classmethod
    def json_schema(cls) -> Optional[dict]:
        """JSON schema for directly-represented enumerations."""
        return {"type": "string", "enum": [item.value for item in cls]}

    @classmethod
    def from_dict(cls, data: str) -> E:
        """Parse an enumeration value from valid JSON data."""
        return cls(data)

    def to_dict(self) -> dict:
        """Return the ready-for-JSON-ification representation."""
        return self.value


class JsonObject(dict):
    """Passing through the json object."""

    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'JsonObject'.

        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object"
        }

    @classmethod
    def from_json(cls, data: dict) -> dict:
        """Validate and parse JSON data into an instance of JsonObject.

        Args:
            data (dict): JSON data to validate and parse.

        Returns:
            Passes the data through.
        """
        return data

    def to_json(self) -> dict:
        """Serialise this instance as JSON.

        Returns:
            JSON data ready to be serialised.
        """
        return self.value


class DataType(dict):
    """Passing through the json object for DataType."""

    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'JsonObject'.

        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object"
        }

    @classmethod
    def from_json(cls, data: dict) -> dict:
        """Validate and parse JSON data into an instance of JsonObject.

        Args:
            data (dict): JSON data to validate and parse.

        Returns:
            Passes the data through.
        """
        return data

    def to_json(self) -> dict:
        """Serialise this instance as JSON.

        Returns:
            JSON data ready to be serialised.
        """
        return self.value



"""Generated implementation of table_caching."""

# WARNING DO NOT EDIT
# This code was generated from table-caching.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..cluster import ClusterId
from ..commit import CommitId
from ..entity import EntityId
from ..feature_store import VersionTarget
from ..feature_store_run import ExecutionStatistics
from ..jobs import RunError, RunStatus
from ..schedule import Schedule
from ..table import TableId


@dataclasses.dataclass(frozen=True)
class TableCachingJobId:
    """Unique identifier of a table caching job.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingJobId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingJobId:
        """Validate and parse JSON data into an instance of TableCachingJobId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingJobId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingJobId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing TableCachingJobId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> TableCachingJobId:
        """Parse a JSON string such as a dictionary key."""
        return TableCachingJobId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class TableCachingJobName:
    """Unique name for a table caching job.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingJobName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingJobName:
        """Validate and parse JSON data into an instance of TableCachingJobName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingJobName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingJobName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing TableCachingJobName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> TableCachingJobName:
        """Parse a JSON string such as a dictionary key."""
        return TableCachingJobName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class TableCachingJobVersionId:
    """Unique identifier for a specific version of a table caching job.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingJobVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingJobVersionId:
        """Validate and parse JSON data into an instance of TableCachingJobVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingJobVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingJobVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing TableCachingJobVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> TableCachingJobVersionId:
        """Parse a JSON string such as a dictionary key."""
        return TableCachingJobVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class TableCachingRunId:
    """Unique identifier of a table caching job run.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingRunId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingRunId:
        """Validate and parse JSON data into an instance of TableCachingRunId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingRunId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingRunId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing TableCachingRunId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return self.value
    
    @classmethod
    def from_json_key(cls, data: str) -> TableCachingRunId:
        """Parse a JSON string such as a dictionary key."""
        return TableCachingRunId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class TableCacheId:
    """Unique identifier of a table cache.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCacheId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCacheId:
        """Validate and parse JSON data into an instance of TableCacheId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCacheId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCacheId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing TableCacheId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> TableCacheId:
        """Parse a JSON string such as a dictionary key."""
        return TableCacheId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class HashBits:
    """The number of bits used to partition the data.
    
    Args:
        bits (int): A data field.
    """
    
    bits: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.bits)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.bits)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for HashBits data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> HashBits:
        """Validate and parse JSON data into an instance of HashBits.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of HashBits.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return HashBits(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing HashBits", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.bits)
    
    @classmethod
    def from_json_key(cls, data: str) -> HashBits:
        """Parse a JSON string such as a dictionary key."""
        return HashBits(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.bits)


@dataclasses.dataclass(frozen=True)
class TableCache:
    """Details of a table cache.
    
    Args:
        id (TableCacheId): A data field.
        job (typing.Optional[TableCachingJobId]): A data field.
        run (typing.Optional[TableCachingRunId]): A data field.
        uri (str): A data field.
        table (TableId): A data field.
        entity (EntityId): A data field.
        hashBits (typing.Optional[HashBits]): A data field.
        commit (CommitId): A data field.
        created (datetime.datetime): A data field.
    """
    
    id: TableCacheId
    job: typing.Optional[TableCachingJobId]
    run: typing.Optional[TableCachingRunId]
    uri: str
    table: TableId
    entity: EntityId
    hashBits: typing.Optional[HashBits]
    commit: CommitId
    created: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCache data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": TableCacheId.json_schema(),
                "job": {
                    "oneOf": [
                        {"type": "null"},
                        TableCachingJobId.json_schema(),
                    ]
                },
                "run": {
                    "oneOf": [
                        {"type": "null"},
                        TableCachingRunId.json_schema(),
                    ]
                },
                "uri": {
                    "type": "string"
                },
                "table": TableId.json_schema(),
                "entity": EntityId.json_schema(),
                "hashBits": {
                    "oneOf": [
                        {"type": "null"},
                        HashBits.json_schema(),
                    ]
                },
                "commit": CommitId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "uri",
                "table",
                "entity",
                "commit",
                "created",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCache:
        """Validate and parse JSON data into an instance of TableCache.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCache.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCache(
                id=TableCacheId.from_json(data["id"]),
                job=(lambda v: v and TableCachingJobId.from_json(v))(data.get("job", None)),
                run=(lambda v: v and TableCachingRunId.from_json(v))(data.get("run", None)),
                uri=str(data["uri"]),
                table=TableId.from_json(data["table"]),
                entity=EntityId.from_json(data["entity"]),
                hashBits=(lambda v: v and HashBits.from_json(v))(data.get("hashBits", None)),
                commit=CommitId.from_json(data["commit"]),
                created=isodate.parse_datetime(data["created"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCache",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "job": (lambda v: v and v.to_json())(self.job),
            "run": (lambda v: v and v.to_json())(self.run),
            "uri": str(self.uri),
            "table": self.table.to_json(),
            "entity": self.entity.to_json(),
            "hashBits": (lambda v: v and v.to_json())(self.hashBits),
            "commit": self.commit.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class TableCachingSpec:
    """Details of the table and entity to be cached.
    
    Args:
        table (TableId): A data field.
        entity (EntityId): A data field.
    """
    
    table: TableId
    entity: EntityId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingSpec data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "table": TableId.json_schema(),
                "entity": EntityId.json_schema()
            },
            "required": [
                "table",
                "entity",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingSpec:
        """Validate and parse JSON data into an instance of TableCachingSpec.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingSpec.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingSpec(
                table=TableId.from_json(data["table"]),
                entity=EntityId.from_json(data["entity"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCachingSpec",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "table": self.table.to_json(),
            "entity": self.entity.to_json()
        }


@dataclasses.dataclass(frozen=True)
class TableCachingJobCreationRequest:
    """Request to create a new table caching job.
    
    Args:
        name (TableCachingJobName): A data field.
        description (str): A data field.
        specs (typing.List[TableCachingSpec]): A data field.
        prefixURI (str): A data field.
        schedule (Schedule): A data field.
        cluster (ClusterId): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
    """
    
    name: TableCachingJobName
    description: str
    specs: typing.List[TableCachingSpec]
    prefixURI: str
    schedule: Schedule
    cluster: ClusterId
    versionTarget: typing.Optional[VersionTarget]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingJobCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": TableCachingJobName.json_schema(),
                "description": {
                    "type": "string"
                },
                "specs": {
                    "type": "array",
                    "item": TableCachingSpec.json_schema()
                },
                "prefixURI": {
                    "type": "string"
                },
                "schedule": Schedule.json_schema(),
                "cluster": ClusterId.json_schema(),
                "versionTarget": {
                    "oneOf": [
                        {"type": "null"},
                        VersionTarget.json_schema(),
                    ]
                }
            },
            "required": [
                "name",
                "description",
                "specs",
                "prefixURI",
                "schedule",
                "cluster",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingJobCreationRequest:
        """Validate and parse JSON data into an instance of TableCachingJobCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingJobCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingJobCreationRequest(
                name=TableCachingJobName.from_json(data["name"]),
                description=str(data["description"]),
                specs=[TableCachingSpec.from_json(v) for v in data["specs"]],
                prefixURI=str(data["prefixURI"]),
                schedule=Schedule.from_json(data["schedule"]),
                cluster=ClusterId.from_json(data["cluster"]),
                versionTarget=(
                    lambda v: v and VersionTarget.from_json(v)
                )(
                    data.get("versionTarget", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCachingJobCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": self.name.to_json(),
            "description": str(self.description),
            "specs": [v.to_json() for v in self.specs],
            "prefixURI": str(self.prefixURI),
            "schedule": self.schedule.to_json(),
            "cluster": self.cluster.to_json(),
            "versionTarget": (lambda v: v and v.to_json())(self.versionTarget)
        }


@dataclasses.dataclass(frozen=True)
class TableCachingJob:
    """Details of a table caching job.
    
    Args:
        id (TableCachingJobId): A data field.
        name (TableCachingJobName): A data field.
        description (str): A data field.
        specs (typing.List[TableCachingSpec]): A data field.
        prefixURI (str): A data field.
        schedule (Schedule): A data field.
        cluster (ClusterId): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
        version (TableCachingJobVersionId): A data field.
    """
    
    id: TableCachingJobId
    name: TableCachingJobName
    description: str
    specs: typing.List[TableCachingSpec]
    prefixURI: str
    schedule: Schedule
    cluster: ClusterId
    versionTarget: typing.Optional[VersionTarget]
    version: TableCachingJobVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingJob data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": TableCachingJobId.json_schema(),
                "name": TableCachingJobName.json_schema(),
                "description": {
                    "type": "string"
                },
                "specs": {
                    "type": "array",
                    "item": TableCachingSpec.json_schema()
                },
                "prefixURI": {
                    "type": "string"
                },
                "schedule": Schedule.json_schema(),
                "cluster": ClusterId.json_schema(),
                "versionTarget": {
                    "oneOf": [
                        {"type": "null"},
                        VersionTarget.json_schema(),
                    ]
                },
                "version": TableCachingJobVersionId.json_schema()
            },
            "required": [
                "id",
                "name",
                "description",
                "specs",
                "prefixURI",
                "schedule",
                "cluster",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingJob:
        """Validate and parse JSON data into an instance of TableCachingJob.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingJob.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingJob(
                id=TableCachingJobId.from_json(data["id"]),
                name=TableCachingJobName.from_json(data["name"]),
                description=str(data["description"]),
                specs=[TableCachingSpec.from_json(v) for v in data["specs"]],
                prefixURI=str(data["prefixURI"]),
                schedule=Schedule.from_json(data["schedule"]),
                cluster=ClusterId.from_json(data["cluster"]),
                versionTarget=(
                    lambda v: v and VersionTarget.from_json(v)
                )(
                    data.get("versionTarget", None)
                ),
                version=TableCachingJobVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCachingJob",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "specs": [v.to_json() for v in self.specs],
            "prefixURI": str(self.prefixURI),
            "schedule": self.schedule.to_json(),
            "cluster": self.cluster.to_json(),
            "versionTarget": (lambda v: v and v.to_json())(self.versionTarget),
            "version": self.version.to_json()
        }


@dataclasses.dataclass(frozen=True)
class TableCachingRunCreationRequest:
    """Request to create a new table caching job run.
    
    Args:
        jobId (TableCachingJobId): A data field.
        jobVersionId (TableCachingJobVersionId): A data field.
        scheduled_start_time (typing.Optional[datetime.datetime]): A data field.
        retryCount (typing.Optional[int]): A data field.
    """
    
    jobId: TableCachingJobId
    jobVersionId: TableCachingJobVersionId
    scheduled_start_time: typing.Optional[datetime.datetime]
    retryCount: typing.Optional[int]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingRunCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "jobId": TableCachingJobId.json_schema(),
                "jobVersionId": TableCachingJobVersionId.json_schema(),
                "scheduled_start_time": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "retryCount": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                }
            },
            "required": [
                "jobId",
                "jobVersionId",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingRunCreationRequest:
        """Validate and parse JSON data into an instance of TableCachingRunCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingRunCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingRunCreationRequest(
                jobId=TableCachingJobId.from_json(data["jobId"]),
                jobVersionId=TableCachingJobVersionId.from_json(data["jobVersionId"]),
                scheduled_start_time=(
                    lambda v: v and isodate.parse_datetime(v)
                )(
                    data.get("scheduled_start_time", None)
                ),
                retryCount=(lambda v: v and int(v))(data.get("retryCount", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCachingRunCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "jobId": self.jobId.to_json(),
            "jobVersionId": self.jobVersionId.to_json(),
            "scheduled_start_time": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.scheduled_start_time),
            "retryCount": (lambda v: v and int(v))(self.retryCount)
        }


@dataclasses.dataclass(frozen=True)
class TableCachingRunUpdateRequest:
    """Request to update a table caching job run.
    
    Args:
        status (RunStatus): A data field.
        error (typing.Optional[RunError]): A data field.
    """
    
    status: RunStatus
    error: typing.Optional[RunError]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingRunUpdateRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "status": RunStatus.json_schema(),
                "error": {
                    "oneOf": [
                        {"type": "null"},
                        RunError.json_schema(),
                    ]
                }
            },
            "required": [
                "status",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingRunUpdateRequest:
        """Validate and parse JSON data into an instance of TableCachingRunUpdateRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingRunUpdateRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingRunUpdateRequest(
                status=RunStatus.from_json(data["status"]),
                error=(lambda v: v and RunError.from_json(v))(data.get("error", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCachingRunUpdateRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "status": self.status.to_json(),
            "error": (lambda v: v and v.to_json())(self.error)
        }


@dataclasses.dataclass(frozen=True)
class TableCachingRun:
    """Details of a table caching job run.
    
    Args:
        id (TableCachingRunId): A data field.
        jobId (TableCachingJobId): A data field.
        jobVersionId (TableCachingJobVersionId): A data field.
        scheduled_start_time (typing.Optional[datetime.datetime]): A data field.
        retryCount (typing.Optional[int]): A data field.
        status (RunStatus): A data field.
        error (typing.Optional[RunError]): A data field.
        executionStatistics (typing.Optional[ExecutionStatistics]): A data field.
        created (datetime.datetime): A data field.
    """
    
    id: TableCachingRunId
    jobId: TableCachingJobId
    jobVersionId: TableCachingJobVersionId
    scheduled_start_time: typing.Optional[datetime.datetime]
    retryCount: typing.Optional[int]
    status: RunStatus
    error: typing.Optional[RunError]
    executionStatistics: typing.Optional[ExecutionStatistics]
    created: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingRun data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": TableCachingRunId.json_schema(),
                "jobId": TableCachingJobId.json_schema(),
                "jobVersionId": TableCachingJobVersionId.json_schema(),
                "scheduled_start_time": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "retryCount": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                },
                "status": RunStatus.json_schema(),
                "error": {
                    "oneOf": [
                        {"type": "null"},
                        RunError.json_schema(),
                    ]
                },
                "executionStatistics": {
                    "oneOf": [
                        {"type": "null"},
                        ExecutionStatistics.json_schema(),
                    ]
                },
                "created": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "jobId",
                "jobVersionId",
                "status",
                "created",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingRun:
        """Validate and parse JSON data into an instance of TableCachingRun.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingRun.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingRun(
                id=TableCachingRunId.from_json(data["id"]),
                jobId=TableCachingJobId.from_json(data["jobId"]),
                jobVersionId=TableCachingJobVersionId.from_json(data["jobVersionId"]),
                scheduled_start_time=(
                    lambda v: v and isodate.parse_datetime(v)
                )(
                    data.get("scheduled_start_time", None)
                ),
                retryCount=(lambda v: v and int(v))(data.get("retryCount", None)),
                status=RunStatus.from_json(data["status"]),
                error=(lambda v: v and RunError.from_json(v))(data.get("error", None)),
                executionStatistics=(
                    lambda v: v and ExecutionStatistics.from_json(v)
                )(
                    data.get("executionStatistics", None)
                ),
                created=isodate.parse_datetime(data["created"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCachingRun",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "jobId": self.jobId.to_json(),
            "jobVersionId": self.jobVersionId.to_json(),
            "scheduled_start_time": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.scheduled_start_time),
            "retryCount": (lambda v: v and int(v))(self.retryCount),
            "status": self.status.to_json(),
            "error": (lambda v: v and v.to_json())(self.error),
            "executionStatistics": (lambda v: v and v.to_json())(self.executionStatistics),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }



"""Generated implementation of merge_request."""

# WARNING DO NOT EDIT
# This code was generated from merge-request.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..anaml_object import AnamlObject
from ..commit import CommitId
from ..user import UserId


@dataclasses.dataclass(frozen=True)
class MergeRequestId:
    """Unique identifier of a merge request.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for MergeRequestId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MergeRequestId:
        """Validate and parse JSON data into an instance of MergeRequestId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequestId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return MergeRequestId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing MergeRequestId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> MergeRequestId:
        """Parse a JSON string such as a dictionary key."""
        return MergeRequestId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class MergeRequestCommentId:
    """Unique identifier of a merge request comment.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for MergeRequestCommentId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MergeRequestCommentId:
        """Validate and parse JSON data into an instance of MergeRequestCommentId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequestCommentId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return MergeRequestCommentId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing MergeRequestCommentId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> MergeRequestCommentId:
        """Parse a JSON string such as a dictionary key."""
        return MergeRequestCommentId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ApprovalId:
    """Unique identifier of a merge request approval.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ApprovalId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ApprovalId:
        """Validate and parse JSON data into an instance of ApprovalId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ApprovalId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ApprovalId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ApprovalId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ApprovalId:
        """Parse a JSON string such as a dictionary key."""
        return ApprovalId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ApprovalVersionId:
    """Unique identifier of a specific version of a merge request approval.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ApprovalVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ApprovalVersionId:
        """Validate and parse JSON data into an instance of ApprovalVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ApprovalVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ApprovalVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ApprovalVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ApprovalVersionId:
        """Parse a JSON string such as a dictionary key."""
        return ApprovalVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


class MergeRequestStatus(enum.Enum):
    """Status of merge requests."""
    Open = "open"
    """The merge request is open."""
    Closed = "closed"
    """The merged request has been closed."""
    Merged = "merged"
    """The merge request has been merged."""
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'MergeRequestStatus'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "open",
                        "closed",
                        "merged",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MergeRequestStatus:
        """Validate and parse JSON data into an instance of MergeRequestStatus.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequestStatus.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return MergeRequestStatus(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing MergeRequestStatus", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> MergeRequestStatus:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequestStatus.
        """
        return MergeRequestStatus(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class MergeRequest:
    """Details of a merge request.
    
    Args:
        id (MergeRequestId): A data field.
        name (str): A data field.
        author (UserId): A data field.
        comment (str): A data field.
        source (str): A data field.
        target (str): A data field.
        status (MergeRequestStatus): A data field.
        mergeCommit (typing.Optional[CommitId]): A data field.
        created (datetime.datetime): A data field.
        modified (datetime.datetime): A data field.
    """
    
    id: MergeRequestId
    name: str
    author: UserId
    comment: str
    source: str
    target: str
    status: MergeRequestStatus
    mergeCommit: typing.Optional[CommitId]
    created: datetime.datetime
    modified: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for MergeRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": MergeRequestId.json_schema(),
                "name": {
                    "type": "string"
                },
                "author": UserId.json_schema(),
                "comment": {
                    "type": "string"
                },
                "source": {
                    "type": "string"
                },
                "target": {
                    "type": "string"
                },
                "status": MergeRequestStatus.json_schema(),
                "mergeCommit": {
                    "oneOf": [
                        {"type": "null"},
                        CommitId.json_schema(),
                    ]
                },
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "modified": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "name",
                "author",
                "comment",
                "source",
                "target",
                "status",
                "created",
                "modified",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MergeRequest:
        """Validate and parse JSON data into an instance of MergeRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return MergeRequest(
                id=MergeRequestId.from_json(data["id"]),
                name=str(data["name"]),
                author=UserId.from_json(data["author"]),
                comment=str(data["comment"]),
                source=str(data["source"]),
                target=str(data["target"]),
                status=MergeRequestStatus.from_json(data["status"]),
                mergeCommit=(
                    lambda v: v and CommitId.from_json(v)
                )(
                    data.get("mergeCommit", None)
                ),
                created=isodate.parse_datetime(data["created"]),
                modified=isodate.parse_datetime(data["modified"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing MergeRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "name": str(self.name),
            "author": self.author.to_json(),
            "comment": str(self.comment),
            "source": str(self.source),
            "target": str(self.target),
            "status": self.status.to_json(),
            "mergeCommit": (lambda v: v and v.to_json())(self.mergeCommit),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "modified": self.modified.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class MergeRequestCreationRequest:
    """Request to create a new merge request.
    
    Args:
        name (str): A data field.
        comment (str): A data field.
        source (str): A data field.
        target (str): A data field.
    """
    
    name: str
    comment: str
    source: str
    target: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for MergeRequestCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string"
                },
                "comment": {
                    "type": "string"
                },
                "source": {
                    "type": "string"
                },
                "target": {
                    "type": "string"
                }
            },
            "required": [
                "name",
                "comment",
                "source",
                "target",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MergeRequestCreationRequest:
        """Validate and parse JSON data into an instance of MergeRequestCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequestCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return MergeRequestCreationRequest(
                name=str(data["name"]),
                comment=str(data["comment"]),
                source=str(data["source"]),
                target=str(data["target"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing MergeRequestCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": str(self.name),
            "comment": str(self.comment),
            "source": str(self.source),
            "target": str(self.target)
        }


@dataclasses.dataclass(frozen=True)
class MergeRequestComment:
    """A new comment on a merge request.
    
    Args:
        id (MergeRequestCommentId): A data field.
        mergeRequest (MergeRequestId): A data field.
        author (UserId): A data field.
        comment (str): A data field.
        created (datetime.datetime): A data field.
        modified (datetime.datetime): A data field.
        anamlObject (typing.Optional[AnamlObject]): A data field.
    """
    
    id: MergeRequestCommentId
    mergeRequest: MergeRequestId
    author: UserId
    comment: str
    created: datetime.datetime
    modified: datetime.datetime
    anamlObject: typing.Optional[AnamlObject]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for MergeRequestComment data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": MergeRequestCommentId.json_schema(),
                "mergeRequest": MergeRequestId.json_schema(),
                "author": UserId.json_schema(),
                "comment": {
                    "type": "string"
                },
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "modified": {
                    "type": "string",
                    "format": "date-time"
                },
                "anamlObject": {
                    "oneOf": [
                        {"type": "null"},
                        AnamlObject.json_schema(),
                    ]
                }
            },
            "required": [
                "id",
                "mergeRequest",
                "author",
                "comment",
                "created",
                "modified",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MergeRequestComment:
        """Validate and parse JSON data into an instance of MergeRequestComment.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequestComment.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return MergeRequestComment(
                id=MergeRequestCommentId.from_json(data["id"]),
                mergeRequest=MergeRequestId.from_json(data["mergeRequest"]),
                author=UserId.from_json(data["author"]),
                comment=str(data["comment"]),
                created=isodate.parse_datetime(data["created"]),
                modified=isodate.parse_datetime(data["modified"]),
                anamlObject=(
                    lambda v: v and AnamlObject.from_json(v)
                )(
                    data.get("anamlObject", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing MergeRequestComment",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "mergeRequest": self.mergeRequest.to_json(),
            "author": self.author.to_json(),
            "comment": str(self.comment),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "modified": self.modified.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "anamlObject": (lambda v: v and v.to_json())(self.anamlObject)
        }


@dataclasses.dataclass(frozen=True)
class MergeRequestCommentCreationRequest:
    """Request to create a new comment on a merge request.
    
    Args:
        comment (str): A data field.
        anamlObject (typing.Optional[AnamlObject]): A data field.
    """
    
    comment: str
    anamlObject: typing.Optional[AnamlObject]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for MergeRequestCommentCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "comment": {
                    "type": "string"
                },
                "anamlObject": {
                    "oneOf": [
                        {"type": "null"},
                        AnamlObject.json_schema(),
                    ]
                }
            },
            "required": [
                "comment",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MergeRequestCommentCreationRequest:
        """Validate and parse JSON data into an instance of MergeRequestCommentCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MergeRequestCommentCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return MergeRequestCommentCreationRequest(
                comment=str(data["comment"]),
                anamlObject=(
                    lambda v: v and AnamlObject.from_json(v)
                )(
                    data.get("anamlObject", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing MergeRequestCommentCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "comment": str(self.comment),
            "anamlObject": (lambda v: v and v.to_json())(self.anamlObject)
        }


@dataclasses.dataclass(frozen=True)
class Approval:
    """Approval of a merge request.
    
    Args:
        id (ApprovalId): A data field.
        version (ApprovalVersionId): A data field.
        predecessor (typing.Optional[ApprovalVersionId]): A data field.
        mergeRequestId (MergeRequestId): A data field.
        approvedBy (UserId): A data field.
        approvalTime (datetime.datetime): A data field.
        approvedCommit (CommitId): A data field.
        comment (typing.Optional[str]): A data field.
    """
    
    id: ApprovalId
    version: ApprovalVersionId
    predecessor: typing.Optional[ApprovalVersionId]
    mergeRequestId: MergeRequestId
    approvedBy: UserId
    approvalTime: datetime.datetime
    approvedCommit: CommitId
    comment: typing.Optional[str]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Approval data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": ApprovalId.json_schema(),
                "version": ApprovalVersionId.json_schema(),
                "predecessor": {
                    "oneOf": [
                        {"type": "null"},
                        ApprovalVersionId.json_schema(),
                    ]
                },
                "mergeRequestId": MergeRequestId.json_schema(),
                "approvedBy": UserId.json_schema(),
                "approvalTime": {
                    "type": "string",
                    "format": "date-time"
                },
                "approvedCommit": CommitId.json_schema(),
                "comment": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                }
            },
            "required": [
                "id",
                "version",
                "mergeRequestId",
                "approvedBy",
                "approvalTime",
                "approvedCommit",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Approval:
        """Validate and parse JSON data into an instance of Approval.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Approval.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Approval(
                id=ApprovalId.from_json(data["id"]),
                version=ApprovalVersionId.from_json(data["version"]),
                predecessor=(
                    lambda v: v and ApprovalVersionId.from_json(v)
                )(
                    data.get("predecessor", None)
                ),
                mergeRequestId=MergeRequestId.from_json(data["mergeRequestId"]),
                approvedBy=UserId.from_json(data["approvedBy"]),
                approvalTime=isodate.parse_datetime(data["approvalTime"]),
                approvedCommit=CommitId.from_json(data["approvedCommit"]),
                comment=(lambda v: v and str(v))(data.get("comment", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Approval",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "version": self.version.to_json(),
            "predecessor": (lambda v: v and v.to_json())(self.predecessor),
            "mergeRequestId": self.mergeRequestId.to_json(),
            "approvedBy": self.approvedBy.to_json(),
            "approvalTime": self.approvalTime.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "approvedCommit": self.approvedCommit.to_json(),
            "comment": (lambda v: v and str(v))(self.comment)
        }


@dataclasses.dataclass(frozen=True)
class ApprovalCreationRequest:
    """Request to create a new approval on a merge request.
    
    Args:
        approvedCommit (CommitId): A data field.
        comment (typing.Optional[str]): A data field.
    """
    
    approvedCommit: CommitId
    comment: typing.Optional[str]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ApprovalCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "approvedCommit": CommitId.json_schema(),
                "comment": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                }
            },
            "required": [
                "approvedCommit",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ApprovalCreationRequest:
        """Validate and parse JSON data into an instance of ApprovalCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ApprovalCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ApprovalCreationRequest(
                approvedCommit=CommitId.from_json(data["approvedCommit"]),
                comment=(lambda v: v and str(v))(data.get("comment", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ApprovalCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "approvedCommit": self.approvedCommit.to_json(),
            "comment": (lambda v: v and str(v))(self.comment)
        }
